# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "puzzlepiece"
copyright = "2024, Jakub Dranczewski"
author = "Jakub Dranczewski"
release = "0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc", "sphinx.ext.viewcode", "sphinx.ext.napoleon",
    "nbsphinx", "sphinx.ext.autosectionlabel"]

templates_path = ["_templates"]
exclude_patterns = []

autodoc_member_order = "bysource"

# autodoc_mock_imports = ["pyqtgraph", "QtPy", "numpy"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
